'use strict'

function isFunction(funktion) {
    return typeof funktion === 'function'
}

function defineProperty(obj, name, value) {
    var enumerable = !!obj[name] && obj.propertyIsEnumerable(name)
    Object.defineProperty(obj, name, {
        configurable: true,
        enumerable: enumerable,
        writable: true,
        value: value
    })
}

var shimmer = function() {}

/**
 * Wrap a single method or multiple methods with the provided wrapper function
 *
 * @param {Object} nodule - The object containing methods to wrap
 * @param {string|Array<string>} name - Method name or array of method names to wrap
 * @param {Function} wrapper - The wrapper function
 * @returns {Function|Array<Function>} - The wrapped function(s)
 */
var wrap = function(nodule, name, wrapper) {
    // If name is an array, recursively call wrap for each name in the array
    if (Array.isArray(name)) {
        var wrapped = [];
        name.forEach(function(methodName) {
            var result = wrap(nodule, methodName, wrapper);
            if (result) wrapped.push(result);
        });
        return wrapped.length > 0 ? wrapped : undefined;
    }

    // Standard single method wrap logic
    if (!nodule || !nodule[name]) {
        return;
    }

    if (!wrapper) {
        return;
    }

    if (!isFunction(nodule[name]) || !isFunction(wrapper)) {
        return;
    }

    var original = nodule[name];
    var wrapped = wrapper(original, name);

    defineProperty(wrapped, '__original', original);
    defineProperty(wrapped, '__unwrap', function() {
        if (nodule[name] === wrapped) defineProperty(nodule, name, original);
    });
    defineProperty(wrapped, '__wrapped', true);

    defineProperty(nodule, name, wrapped);
    return wrapped;
}

/**
 * Unwrap a single method or multiple methods
 *
 * @param {Object} nodule - The object containing wrapped methods
 * @param {string|Array<string>} name - Method name or array of method names to unwrap
 */
function unwrap(nodule, name) {
    // If name is an array, recursively call unwrap for each name in the array
    if (Array.isArray(name)) {
        name.forEach(function(methodName) {
            unwrap(nodule, methodName);
        });
        return;
    }

    if (!nodule || !nodule[name]) {
        return;
    }

    if (!nodule[name].__unwrap) {
        // Method is not wrapped
    } else {
        return nodule[name].__unwrap();
    }
}

/**
 * Wrap multiple methods of multiple modules with the same wrapper
 *
 * @param {Object|Array<Object>} nodules - Module or array of modules with methods to wrap
 * @param {Array<string>} names - Array of method names to wrap
 * @param {Function} wrapper - The wrapper function to apply
 */
function massWrap(nodules, names, wrapper) {
    if (!nodules) {
        return;
    } else if (!Array.isArray(nodules)) {
        nodules = [nodules];
    }

    if (!(names && Array.isArray(names))) {
        return;
    }

    nodules.forEach(function(nodule) {
        names.forEach(function(name) {
            wrap(nodule, name, wrapper);
        });
    });
}

/**
 * Unwrap multiple methods of multiple modules
 *
 * @param {Object|Array<Object>} nodules - Module or array of modules with wrapped methods
 * @param {Array<string>} names - Array of method names to unwrap
 */
function massUnwrap(nodules, names) {
    if (!nodules) {
        return;
    } else if (!Array.isArray(nodules)) {
        nodules = [nodules];
    }

    if (!(names && Array.isArray(names))) {
        return;
    }

    nodules.forEach(function(nodule) {
        names.forEach(function(name) {
            unwrap(nodule, name);
        });
    });
}

shimmer.wrap = wrap;
shimmer.unwrap = unwrap;
shimmer.massWrap = massWrap;
shimmer.massUnwrap = massUnwrap;

module.exports = shimmer;