/**
 * Copyright 2025 WHATAP project authors. All rights reserved.
 * Use of this source code is governed by a license that
 * can be found in the LICENSE file.
 */

const TraceContextManager = require("../trace/trace-context-manager");
const ParsedSql = require("../trace/parsed-sql");
const SqlStepX = require("../step/sql-stepx");
const DBCStep = require("../step/dbc-step");
const ResultSetStep = require("../step/resultset-step");
const DataTextAgent = require("../data/datatext-agent");
const StatSql = require("../stat/stat-sql");
const MeterSql = require("../counter/meter/meter-sql");
const conf = require("../conf/configure");
const IntKeyMap = require("../util/intkey-map");
const EscapeLiteralSQL = require("../util/escape-literal-sql");
const HashUtil = require("../util/hashutil");
const StatError = require("../stat/stat-error");
const TextTypes = require("../lang/text-types");
const ParamSecurity = require("../util/paramsecurity");
const Logger = require("../logger");
const Buffer = require("buffer").Buffer;
const DateUtil = require("../util/dateutil");
const TraceSQL = require("../trace/trace-sql");
const shimmer = require("../core/shimmer");

var prisma_read_func_name = conf.getProperty('prisma_read_func_name', 'read');
var prisma_database_url_name = conf.getProperty('prisma_database_url_name', 'DATABASE_URL')
conf.on('prisma_database_url_name', function (newProps) {
    prisma_database_url_name = newProps;
})

var PrismaObserver = function(agent) {
    this.agent = agent;
    this.packages = ["@prisma/client"];
};

var dbc_hash = 0;
var dbc = "";

PrismaObserver.prototype.inject = function(mod, moduleName) {
    if (mod.__whatap_observe__) {
        return;
    }

    mod.__whatap_observe__ = true;
    Logger.initPrint("PrismaObserver");

    const self = this;

    if (conf.sql_enabled === false) {
        return;
    }

    // Prisma Client 초기화 메서드 후킹
    if (mod.PrismaClient) {
        // 직접 PrismaClient 생성자 후킹
        shimmer.wrap(mod, 'PrismaClient', function(originalConstructor) {
            return function(...args) {
                const originalInstance = new originalConstructor(...args);
                const instance = Object.create(originalInstance);

                const prismaServicePrototype = Object.getPrototypeOf(this);
                Object.getOwnPropertyNames(prismaServicePrototype).forEach((method) => {
                    if (typeof prismaServicePrototype[method] === "function" && !instance[method]) {
                        instance[method] = prismaServicePrototype[method].bind(instance);
                    }
                });

                Object.getOwnPropertyNames(originalInstance).forEach((prop) => {
                    if (!instance.hasOwnProperty(prop)) {
                        instance[prop] = originalInstance[prop];
                    }
                });

                self.patchPrismaInstance(instance);

                if (!instance[prisma_read_func_name]) {
                    instance[prisma_read_func_name] = function () {
                        return this;
                    };
                }

                return instance;
            };
        });
    }
};

// 각 Prisma 인스턴스에 패치 적용
PrismaObserver.prototype.patchPrismaInstance = function(prismaInstance) {
    if (prismaInstance.__whatap_observe__) {
        return;
    }
    prismaInstance.__whatap_observe__ = true;

    this.setupConnectionInfo(prismaInstance);

    this.hookUseMiddleware(prismaInstance);
};

// 연결 정보 설정
PrismaObserver.prototype.setupConnectionInfo = function(prismaInstance) {
    try {
        // 연결 정보 가져오기 시도
        const url = prismaInstance._engineConfig?.overrideDatasources?.db?.url ||
            prismaInstance.env?.DATABASE_URL ||
            process.env[prisma_database_url_name] ||
            'prisma:unknown';

        if (url && url !== "prisma:unknown" && !dbc_hash) {
            const dbUrl = new URL(url);
            const protocol = dbUrl.protocol.replace(':', '');

            // 프로토콜에 따라 접두사 설정 (postgresql -> pgsql로 변환할 수도 있음)
            const dbProtocol = protocol === 'pgsql' ? 'postgresql' : protocol;

            // MySQL 관찰자와 동일한 형식으로 구성
            dbc = `${dbProtocol}://`;
            dbc += dbUrl.username || '';
            dbc += "@";
            dbc += dbUrl.hostname || '';
            dbc += '/';
            dbc += dbUrl.pathname.replace('/', '') || '';
            dbc_hash = HashUtil.hashFromString(dbc);

            DataTextAgent.DBC.add(dbc_hash, dbc);
            DataTextAgent.METHOD.add(dbc_hash, dbc);
            DataTextAgent.ERROR.add(dbc_hash, dbc);
        }

    } catch (e) {
        Logger.printError("WHATAP-301", "Failed to extract connection info", e, false);
        dbc = "prisma:unknown";
        dbc_hash = HashUtil.hashFromString(dbc);
    }
};

PrismaObserver.prototype.hookUseMiddleware = function(prismaInstance) {
    const originalUse = prismaInstance.$use;
    if (typeof originalUse === 'function') {
        prismaInstance.$use(async (params, next) => {
            var result;
            const ctx = TraceContextManager.getCurrentContext();
            if (!ctx) {
                return next(params);
            }

            var dbc_step = new DBCStep();
            dbc_hash = HashUtil.hashFromString(dbc);
            DataTextAgent.DBC.add(dbc_hash, dbc);
            DataTextAgent.METHOD.add(dbc_hash, dbc);
            DataTextAgent.ERROR.add(dbc_hash, dbc)

            dbc_step.hash = dbc_hash;
            dbc_step.start_time = ctx.getElapsedTime();
            ctx.profile.push(dbc_step);

            const sql_step = new SqlStepX();
            sql_step.start_time = ctx.getElapsedTime();
            ctx.profile.push(sql_step);

            const modelName = params.model || 'unknown';
            const action = params.action || 'unknown';

            ctx.footprint(`Prisma ${modelName}.${action} Start`);
            ctx.sql_count = (ctx.sql_count || 0) + 1;

            // Prisma 쿼리 정보 직접 사용 (SQL 변환 없이)
            const queryInfo = `Prisma ${modelName}.${action}`;
            const queryHash = HashUtil.hashFromString(queryInfo);

            DataTextAgent.SQL.add(queryHash, queryInfo);
            sql_step.hash = queryHash;
            sql_step.dbc = dbc_hash;

            ctx.active_sqlhash = sql_step.hash;
            ctx.active_dbc = sql_step.dbc;

            // 쿼리 파라미터 정보 추출
            if (conf.profile_sql_param_enabled) {
                const paramsString = JSON.stringify(params.args || {});
                sql_step.setTrue(1);
                var crc = { value: 0 };
                sql_step.p1 = toParamBytes(paramsString, crc);
                sql_step.pcrc = crc.value;
            }

            try {
                result = await next(params);

                // Raw 쿼리 처리
                if (action === "queryRaw" || action === "executeRaw" || action === "queryRawUnsafe" || action === "executeRawUnsafe") {
                    // SQL 문자열 추출
                    let sqlString = "";
                    if (params.args && params.args.length > 0) {
                        // 첫 번째 인자가 문자열 배열인 경우 (tagged template)
                        if (params.args[0] && params.args[0].strings && Array.isArray(params.args[0].strings)) {
                            sqlString = params.args[0].strings.join('');
                        }
                        // 첫 번째 인자가 직접 문자열인 경우
                        else if (typeof params.args[0] === 'string') {
                            sqlString = params.args[0];
                        }
                    }

                    // SQL 문자열이 있으면 처리
                    if (sqlString && sqlString.length > 0) {
                        try {
                            var psql = escapeLiteral(sqlString);
                            if (psql != null) {
                                sql_step.hash = psql.sql;
                            }

                            // 추가 SQL 정보 처리
                            var els = new EscapeLiteralSQL(sqlString);
                            els.process();

                            // SQL 파라미터 처리
                            if (conf.profile_sql_param_enabled) {
                                var params = params.args.slice(1);
                                sql_step.setTrue(1);
                                var crc = {value: 0};
                                sql_step.p1 = toParamBytes(psql.param, crc);

                                if (params && params.length > 0) {
                                    const result = params.map((param) => {
                                        if (typeof param === 'string') {
                                            return `'${param}'`;
                                        }
                                        return param;
                                    }).toString();
                                    sql_step.p2 = toParamBytes(result, crc);
                                }
                                sql_step.pcrc = crc.value;
                            }
                        } catch (e) {
                            Logger.printError("WHATAP-305", "escapeLiteral error", e, false);
                        }
                    }

                    // Raw 쿼리 결과셋 처리
                    if ((action === "queryRaw" || action === "queryRawUnsafe") && result) {
                        let recordCount = 0;

                        if (Array.isArray(result)) {
                            recordCount = result.length;
                        } else if (result && typeof result === "object") {
                            recordCount = 1;
                        }

                        if(psql && psql.sql){
                            const sqlHashToUse = psql.sql;

                            var result_step = new ResultSetStep();
                            result_step.start_time = ctx.getElapsedTime();
                            result_step.elapsed = 0;
                            result_step.fetch = recordCount;
                            result_step.sqlhash = sqlHashToUse;
                            result_step.dbc = dbc_hash;
                            ctx.profile.push(result_step);

                            ctx.rs_count = ctx.rs_count ? ctx.rs_count + recordCount : recordCount;
                            ctx.rs_time = ctx.rs_time ? ctx.rs_time + sql_step.elapsed : sql_step.elapsed;

                            MeterSql.addFetch(result_step.dbc, result_step.fetch, 0);
                            StatSql.addFetch(result_step.dbc, result_step.sqlhash, result_step.fetch, 0);

                            TraceSQL.isTooManyRecords(sql_step, result_step.fetch, ctx);
                        }
                    }
                }

                // 결과셋 처리 (findMany, findFirst, findUnique 등)
                if (["findMany", "findFirst", "findUnique"].includes(action)) {
                    let recordCount = 0;

                    if (Array.isArray(result)) {
                        recordCount = result.length;
                    } else if (result && typeof result === "object") {
                        recordCount = 1;
                    }

                    var result_step = new ResultSetStep();
                    result_step.start_time = ctx.getElapsedTime();
                    result_step.elapsed = 0;
                    result_step.fetch = recordCount;
                    result_step.sqlhash = queryHash;
                    result_step.dbc = dbc_hash;
                    ctx.profile.push(result_step);

                    ctx.rs_count = ctx.rs_count ? ctx.rs_count + recordCount : recordCount;
                    ctx.rs_time = ctx.rs_time ? ctx.rs_time + sql_step.elapsed : sql_step.elapsed;

                    MeterSql.addFetch(result_step.dbc, result_step.fetch, 0);
                    StatSql.addFetch(result_step.dbc, result_step.sqlhash, result_step.fetch, 0);

                    TraceSQL.isTooManyRecords(sql_step, result_step.fetch, ctx);
                }

                // 수정된 레코드 수 처리 (create, update, delete 등)
                if (["create", "createMany", "update", "updateMany", "delete", "deleteMany"].includes(action)) {
                    if (result && result.count !== undefined) {
                        sql_step.updated = result.count;
                    } else if (result && typeof result === "object") {
                        sql_step.updated = 1;
                    }
                }

                // UPSERT 처리 추가
                if (action === "upsert") {
                    sql_step.updated = 1; // upsert는 항상 1개의 레코드에 영향을 미침
                }

                this._finishQuery(ctx, sql_step);

                ctx.footprint(`Prisma ${modelName}.${action} Done`);

                return result;
            } catch (err) {
                Logger.printError("WHATAP-308", `Middleware error in ${modelName}.${action}: ${err.message}`, err, false);

                this._handleError(ctx, sql_step, err);
                ctx.footprint(`Prisma ${modelName}.${action} Error`);
                throw err;
            }
        });
    }
};

PrismaObserver.prototype._finishQuery = function(ctx, sql_step) {
    sql_step.elapsed = ctx.getElapsedTime() - sql_step.start_time;
    ctx.sql_time = (ctx.sql_time || 0) + sql_step.elapsed;

    TraceSQL.isSlowSQL(ctx);

    MeterSql.add(dbc_hash, sql_step.elapsed, false);
    StatSql.addSqlTime(ctx.service_hash, sql_step.dbc, sql_step.hash, sql_step.elapsed, false, 0);

    ctx.footprint("Prisma Query Done");
};

PrismaObserver.prototype._handleError = function(ctx, sql_step, err) {
    sql_step.elapsed = ctx.getElapsedTime() - sql_step.start_time;
    ctx.sql_time = (ctx.sql_time || 0) + sql_step.elapsed;

    TraceSQL.isSlowSQL(ctx);

    MeterSql.add(dbc_hash, sql_step.elapsed, true);
    StatSql.addSqlTime(ctx.service_hash, sql_step.dbc, sql_step.hash, sql_step.elapsed, true, 0);

    try {
        sql_step.error = StatError.addError("prisma-" + (err.code || "unknown"), err.message, ctx.service_hash, TextTypes.SQL, sql_step.hash);

        if (ctx.error.isZero()) {
            ctx.error = sql_step.error;
        }

        if (conf.trace_sql_error_stack && conf.trace_sql_error_depth) {
            var traceDepth = conf.trace_sql_error_depth;
            var errorStack = err.stack.split("\n");

            if (errorStack.length > traceDepth) {
                errorStack = errorStack.slice(0, traceDepth + 1);
            }
            ctx.error_message = errorStack.join("\n");
        }
    } catch (e) {
        Logger.printError("WHATAP-309", "Error handling failed", e, false);
    }

    ctx.footprint("Prisma Query Error");
};

var toParamBytes = function(p, crc) {
    if (p == null || p.length === 0) {
        return null;
    }
    try {
        return ParamSecurity.encrypt(Buffer.from(p, "utf8"), crc);
    } catch (e) {
        return null;
    }
};

var checkedSql = new IntKeyMap(2000).setMax(2000);
var nonLiteSql = new IntKeyMap(5000).setMax(5000);
var date = DateUtil.yyyymmdd();

function escapeLiteral(sql) {
    if (sql == null) {
        sql = "";
    }

    if (date !== DateUtil.yyyymmdd()) {
        checkedSql.clear();
        nonLiteSql.clear();
        date = DateUtil.yyyymmdd();
        Logger.print("WHATAP-SQL-CLEAR", "PrismaObserver CLEAR OK!!!!!!!!!!!!!!!!", false);
    }

    var sqlHash = HashUtil.hashFromString(sql);
    var psql = nonLiteSql.get(sqlHash);

    if (psql != null) {
        return psql;
    }
    psql = checkedSql.get(sqlHash);

    if (psql != null) {
        return psql;
    }

    var els = new EscapeLiteralSQL(sql);
    els.process();

    var hash = HashUtil.hashFromString(els.getParsedSql());
    DataTextAgent.SQL.add(hash, els.getParsedSql());

    if (hash === sqlHash) {
        psql = new ParsedSql(els.sqlType, hash, null);
        nonLiteSql.put(sqlHash, psql);
    } else {
        psql = new ParsedSql(els.sqlType, hash, els.getParameter());
        checkedSql.put(sqlHash, psql);
    }
    return psql;
}

exports.PrismaObserver = PrismaObserver;