var TraceContextManager = require('../trace/trace-context-manager'),
    DataTextAgent = require('../data/datatext-agent'),
    SqlStepX = require('../step/sql-stepx'),
    DBCStep = require('../step/dbc-step'),
    HashUtil = require('../util/hashutil'),
    Logger = require('../logger'),
    conf = require('../conf/configure'),
    ParamSecurity = require('../util/paramsecurity');
const shimmer = require('../core/shimmer');
const {Buffer} = require("buffer");

var RedisObserver = function (agent) {
    this.agent = agent;
    this.packages = ['redis', 'ioredis'];
};

RedisObserver.prototype.inject = function (mod, modName) {
    if (mod.__whatap_observe__) {
        return;
    }
    mod.__whatap_observe__ = true;
    Logger.initPrint("RedisObserver");

    if (modName === 'redis') {
        this._injectRedis(mod);
    } else if (modName === 'ioredis') {
        this._injectIORedis(mod);
    }
};

RedisObserver.prototype._injectRedis = function (mod) {
    var dbc_hash = 0;
    var dbc = '';

    const selectCommand = new Set(['get', 'hGet', 'hmGet']);
    const insertCommand = new Set(['set', 'hSet', 'hset', 'hmSet', 'hmset', 'zAdd', 'zadd', 'lSet', 'lset']);
    const updateCommand = new Set(['set', 'lSet', 'lset', 'hSet', 'hset', 'zAdd', 'zadd']);
    const deleteCommand = new Set(['del', 'lRem', 'sRem', 'srem', 'hDel', 'hdel', 'zRem', 'zrem']);
    const commands = new Set([...selectCommand, ...insertCommand, ...updateCommand, ...deleteCommand]);

    shimmer.wrap(mod, 'createClient', function (original) {
        return function wrappedCreateClient() {
            if (dbc_hash === 0) {
                if (arguments.length > 0) {
                    var info = (arguments[0] || {});
                    dbc = info.url || '';
                    dbc_hash = HashUtil.hashFromString(dbc);

                    DataTextAgent.DBC.add(dbc_hash, dbc);
                    DataTextAgent.METHOD.add(dbc_hash, dbc);
                    DataTextAgent.ERROR.add(dbc_hash, dbc);
                }
            }

            var client = original.apply(this, arguments);

            const commandsArray = Array.from(commands);
            commandsArray.forEach(cmdName => {
                shimmer.wrap(client, cmdName, function (originalMethod) {
                    const wrappedFunction = function wrappedSendCommand() {
                        try{
                            var ctx = TraceContextManager.getCurrentContext();
                            if (!ctx) {
                                return originalMethod.apply(this, arguments);
                            }

                            var dbc_step = new DBCStep();
                            dbc_step.hash = dbc_hash;
                            ctx.profile.push(dbc_step);

                            ctx.footprint('Redis Command Start');

                            var command = wrappedFunction._commandName || 'unknown';
                            var sql = `${command.toUpperCase()}`;
                            var sql_step = new SqlStepX();
                            sql_step.start_time = ctx.getElapsedTime();
                            sql_step.dbc = dbc_hash;

                            if(conf.getProperty('profile_redis_param_enabled', false)){
                                if (arguments[0] && typeof arguments[0] === 'string')
                                    sql += ' ' + arguments[0];
                                else if (arguments[0] && Array.isArray(arguments[0]))
                                    sql += ' ' + JSON.stringify(arguments[0]);
                                else if (arguments[0] && typeof arguments[0] === 'object'){
                                    sql += ' ' + JSON.stringify(Object.keys(arguments[0]))
                                }

                                if(arguments[1] && arguments[1].length > 0){
                                    var args = Array.isArray(arguments[1]) ? arguments[1] : [arguments[1]];
                                    sql_step.setTrue(1);
                                    var crc = {value: 0};
                                    const result = args.map((arg) => {
                                        if (typeof arg === 'string') {
                                            return `'${arg}'`
                                        }
                                        return arg
                                    }).toString()
                                    sql_step.p2 = toParamBytes(result, crc);
                                    sql_step.pcrc = crc.value;
                                }
                            }
                            sql_step.hash = HashUtil.hashFromString(sql);
                            ctx.profile.push(sql_step);

                            DataTextAgent.SQL.add(sql_step.hash, sql);

                            var result = originalMethod.apply(this, arguments);
                            sql_step.elapsed = ctx.getElapsedTime();

                            ctx.footprint('Redis Command Done');
                            return result;
                        }catch (e) {
                            Logger.printError("WHATAP-605", "Redis CRUD error", e, false);
                            return originalMethod.apply(this, arguments);
                        }
                    };
                    wrappedFunction._commandName = cmdName;
                    return wrappedFunction;
                });
            });
            return client;
        };
    });

    return this;
};

RedisObserver.prototype._injectIORedis = function (mod) {
    var ioredis_dbc_hash = 0;
    var ioredis_dbc = '';

    const selectCommand = new Set(['get', 'hget', 'hmget', 'zrange', 'smembers', 'lrange']);
    const insertCommand = new Set(['set', 'hset', 'hmset', 'zadd', 'lset', 'sadd', 'lpush', 'rpush', 'evalsha']);
    const updateCommand = new Set(['set', 'lset', 'hset', 'zadd']);
    const deleteCommand = new Set(['del', 'lrem', 'srem', 'hdel', 'zrem']);
    const commands = new Set([...selectCommand, ...insertCommand, ...updateCommand, ...deleteCommand]);

    shimmer.wrap(mod.prototype, 'sendCommand', function (original) {
        return function wrappedSendCommand(command) {
            // Get trace context
            var ctx = TraceContextManager.getCurrentContext();
            if (!ctx) {
                return original.apply(this, arguments);
            }

            if (ioredis_dbc_hash === 0 && this.options) {
                ioredis_dbc = 'redis://';
                ioredis_dbc += this.options.host || '';
                ioredis_dbc += ':';
                ioredis_dbc += this.options.port || '';
                ioredis_dbc_hash = HashUtil.hashFromString(ioredis_dbc);
                DataTextAgent.DBC.add(ioredis_dbc_hash, ioredis_dbc);
            }

            if (command && command.name && commands.has(command.name.toLowerCase())) {
                var dbc_step = new DBCStep();
                dbc_step.hash = ioredis_dbc_hash;
                dbc_step.start_time = ctx.getElapsedTime();
                ctx.profile.push(dbc_step);

                try {
                    var sql_step = new SqlStepX();
                    var sql = command.name.toUpperCase();
                    if(conf.getProperty('profile_redis_param_enabled', false)){
                        var args = command.args;
                        var key = args.shift();
                        if (key && typeof key === 'string')
                            sql += ' ' + key;

                        if(args && args.length > 0){
                            sql_step.setTrue(1);
                            var crc = {value: 0};
                            const result = args.map((arg) => {
                                if (typeof arg === 'string') {
                                    return `'${arg}'`
                                }
                                return arg;
                            }).toString()
                            sql_step.p2 = toParamBytes(result, crc);
                            sql_step.pcrc = crc.value;
                        }
                    }

                    sql_step.hash = HashUtil.hashFromString(sql);
                    sql_step.start_time = ctx.getElapsedTime();
                    DataTextAgent.SQL.add(sql_step.hash, sql);
                    ctx.profile.push(sql_step);

                    var originalPromise = original.apply(this, arguments);
                    return originalPromise.finally(function () {
                        dbc_step.elapsed = ctx.getElapsedTime() - dbc_step.start_time;
                        sql_step.elapsed = ctx.getElapsedTime() - sql_step.start_time;
                    });
                } catch (e) {
                    Logger.printError("WHATAP-605", "ioredis CRUD error", e, false);
                    return original.apply(this, arguments);
                }
            }

            return original.apply(this, arguments);
        };
    });

    return this;
};

var toParamBytes = function (p, crc) {
    if (p == null || p.length === 0) {
        return null;
    }
    try {
        return ParamSecurity.encrypt(Buffer.from(p, 'utf8'), crc);
    } catch (e) {
        return null;
    }
};

exports.RedisObserver = RedisObserver;