/**
 * Copyright 2016 the WHATAP project authors. All rights reserved.
 * Use of this source code is governed by a license that
 * can be found in the LICENSE file.
 */


var TraceContextManager = require('../trace/trace-context-manager'),
    SocketStep = require('../step/socket-step'),
    conf = require('../conf/configure'),
    IPUtil = require('../util/iputil'),
    Logger = require('../logger');
const {Detector: URLPatternDetector} = require("../trace/serviceurl-pattern-detector");
const HashUtil = require("../util/hashutil");
const DataTextAgent = require("../data/datatext-agent");
const ResourceProfile = require("../util/resourceprofile");
const ProfilePack         = require('../pack/profile-pack');
const TxRecord            = require('../service/tx-record');
const DateUtil            = require('../util/dateutil');
const SecurityMaster      = require('../net/security-master');
const DataProfileAgent    = require('../data/dataprofile-agent');
const MessageStep = require("../step/message-step");
const MeterUsers = require("../counter/meter/meter-users");
const MeterService        = require('../counter/meter/meter-service').MeterService;
const shimmer = require('../core/shimmer');

// 설정을 객체로 통합하여 관리 (참조 성능 향상)
var config = {
    trace_background_socket_enabled: conf.getProperty('trace_background_socket_enabled', true),
    trace_sampling_enabled: conf.getProperty('trace_sampling_enabled', true),
    trace_sampling_tps: conf.getProperty('trace_sampling_tps', 1000),
    resource_sampling_rate: 0.1,
    profile_batch_size: 20,
    flush_interval: 500
};

conf.on('trace_background_socket_enabled', function (newProps) {
    config.trace_background_socket_enabled = newProps;
});

conf.on('trace_sampling_enabled', function (newProps) {
    config.trace_sampling_enabled = newProps;
});

conf.on('trace_sampling_tps', function (newProps) {
    config.trace_sampling_tps = newProps;
});

var WebsocketObserver = function(agent){
    this.agent = agent;
    this.packages = ['websocket'];

    this.socketCounter = {
        count: 0,
        start_time: Date.now(),
        window_size: 5000,

        checkSampling: function() {
            const now = Date.now();
            if ((now - this.start_time) >= this.window_size) {
                this.start_time = now;
                this.count = 0;
                return true;
            }

            this.count++;
            return this.count <= config.trace_sampling_tps;
        }
    };

    this.profileBuffer = [];

    // 주기적 플러시 설정
    setInterval(() => this.flushProfileBuffer(), config.flush_interval);

    this.ipCache = new Map();
};

WebsocketObserver.prototype.flushProfileBuffer = function() {
    if (this.profileBuffer.length === 0) return;

    const now = Date.now();
    // 100ms 이상 지난 항목만 처리
    const bufferToFlush = this.profileBuffer.filter(item => (now - item.timestamp) >= 100);
    if (bufferToFlush.length === 0) return;

    // 최대 batch_size까지만 처리
    let batchToProcess;
    if(bufferToFlush.length > config.profile_batch_size)
        batchToProcess = bufferToFlush.slice(0, config.profile_batch_size);
    else
        batchToProcess = bufferToFlush;

    // 버퍼에서 처리할 항목 제거
    this.profileBuffer = this.profileBuffer.filter(item =>
        !batchToProcess.some(batch => batch.ctx._id === item.ctx._id)
    );

    // 배치 처리
    batchToProcess.forEach(item => {
        try {
            DataProfileAgent.sendProfile(item.ctx, item.profile, false);
            // 컨텍스트 정리 (메모리 누수 방지)
            item.ctx = null;
        } catch (e) {
            Logger.printError('WHATAP-615', 'Websocket buffer flush error', e, false);
        }
    });

    if (this.profileBuffer.length > 1000) {
        // 가장 오래된 항목부터 제거
        this.profileBuffer = this.profileBuffer.slice(-500);
    }
};

WebsocketObserver.prototype.inject = function (mod, moduleName) {
    if (mod.__whatap_observe__) {
        return;
    }
    mod.__whatap_observe__ = true;
    Logger.initPrint("WebsocketObserver");

    var self = this;
    var aop = self.agent.aop;

    // IP 주소 변환 캐시 (동일한 IP 변환 작업 최소화)
    const ipCache = new Map();

    function getProcessedIp(remoteAddress) {
        if (!remoteAddress) return null;

        if (ipCache.has(remoteAddress)) {
            return ipCache.get(remoteAddress);
        }

        let host = remoteAddress;
        if (remoteAddress.includes(':')) {
            host = remoteAddress.substring(remoteAddress.lastIndexOf(':') + 1);
        }

        host = IPUtil.checkIp4(host);
        const ipInt = IPUtil.stringToInt(host);
        const ipBytes = Buffer.from(IPUtil.stringToBytes(host));

        const result = { host, ipInt, ipBytes };
        ipCache.set(remoteAddress, result);

        return result;
    }

    aop.after(mod.server.prototype, 'on', function (obj, args, ret, lctx){
        aop.after(mod.connection.prototype, ['send', 'sendUTF', 'sendBytes'], function (obj, args, ret, lctx) {
            if(!config.trace_background_socket_enabled) {
                return;
            }

            if(config.trace_sampling_enabled) {
                if(!self.socketCounter.checkSampling()) {
                    MeterService.add(0, 1, 0, 0, 0, 0);
                    return;
                }
            }

            var protocol = obj.protocol;
            var remoteAddress = obj.remoteAddress;

            // 리소스 프로파일링 최적화 - 모든 트랜잭션 대신 샘플링
            const shouldProfileResource = Math.random() < config.resource_sampling_rate; // 10%만 리소스 프로파일링

            TraceContextManager._asyncLocalStorage.run(initCtx(obj, args, shouldProfileResource), () => {
                try {
                    var ctx = TraceContextManager._asyncLocalStorage.getStore();
                    if (!ctx) {
                        return;
                    }

                    // IP 주소 처리 최적화
                    const ipInfo = getProcessedIp(remoteAddress);
                    if (!ipInfo) return;

                    // 불필요한 footprint 로깅 제거 (메모리 사용 감소)
                    ctx.socket_connecting = true;

                    // 프로토콜 정보 수집 최적화
                    if (protocol) {
                        var protocol_hash = HashUtil.hashFromString('Protocol');
                        var step = new MessageStep();
                        step.hash = protocol_hash;
                        step.start_time = ctx.getElapsedTime();
                        step.desc = protocol;

                        DataTextAgent.MESSAGE.add(protocol_hash, 'Protocol');
                        ctx.profile.add(step);
                    }

                    // 소켓 단계 기록 (메모리 사용 최적화)
                    var step = new SocketStep();
                    step.start_time = ctx.getElapsedTime();
                    step.ipaddr = ipInfo.ipBytes;

                    ctx.socket_connecting = false;
                    step.elapsed = ctx.getElapsedTime() - step.start_time;
                    ctx.profile.push(step);

                    // 트랜잭션 종료 처리 최적화
                    self.__endTransaction(null, ctx);
                    return;
                } catch (e) {
                    Logger.printError('WHATAP-616', 'Websocket transaction error', e, false);
                    return;
                }
            });
        });
    });
};

WebsocketObserver.prototype.__endTransaction = function(error, ctx) {
    try {
        var profile = new ProfilePack();
        var wtx = new TxRecord();
        wtx.endTime = DateUtil.currentTime();
        profile.time = wtx.endTime;
        wtx.elapsed = ctx.getElapsedTime();

        DataTextAgent.SERVICE.add(ctx.service_hash, ctx.service_name);

        wtx.seq = ctx.txid;
        wtx.service = ctx.service_hash;

        // CPU와 메모리 측정은 샘플링된 트랜잭션에만 수행
        if (ctx.shouldProfileResource) {
            wtx.cpuTime = ResourceProfile.getCPUTime() - ctx.start_cpu;
            wtx.malloc = ResourceProfile.getUsedHeapSize() - ctx.start_malloc;
            if(wtx.malloc < 0) { wtx.malloc = 0; }
        } else {
            wtx.cpuTime = 0;
            wtx.malloc = 0;
        }

        wtx.status = 2;
        wtx.ipaddr = ctx.remoteIp;

        MeterService.add(0, wtx.elapsed, 0, SecurityMaster.PCODE, SecurityMaster.OKIND, SecurityMaster.OID);

        profile.oid = SecurityMaster.OID;
        profile.service = wtx;

        // 트랜잭션 컨텍스트 종료 (메모리 누수 방지)
        TraceContextManager.end(ctx._id);

        // 프로파일 데이터 버퍼에 추가 (비동기 처리 최적화)
        this.profileBuffer.push({
            ctx: ctx,
            profile: profile,
            timestamp: Date.now()
        });

    } catch (e) {
        Logger.printError('WHATAP-615', 'Websocket end transaction error', e, false);
        // 에러 발생 시에도 컨텍스트 정리
        TraceContextManager.end(ctx._id);
        ctx = null;
    }
};

// 컨텍스트 초기화 최적화
function initCtx(socket, args, shouldProfileResource) {
    const ctx = TraceContextManager.start();
    if (!ctx) {return;}

    var remote_addr;
    const address = socket.remoteAddress;
    if(address && address.includes(':')){
        remote_addr = address.substring(address.lastIndexOf(':')+1);
    }

    // 리소스 측정은 샘플링된 트랜잭션에만 수행 (CPU 사용 감소)
    ctx.shouldProfileResource = shouldProfileResource;
    if (shouldProfileResource) {
        ctx.start_malloc = ResourceProfile.getUsedHeapSize();
        ctx.start_cpu = ResourceProfile.getCPUTime();
    }

    remote_addr = IPUtil.checkIp4(remote_addr);
    ctx.remoteIp = IPUtil.stringToInt(remote_addr);
    ctx.userid = Long.fromNumber(ctx.remoteIp);
    MeterUsers.add(ctx.userid);

    return ctx;
}

exports.WebsocketObserver = WebsocketObserver;